/*

Filename : EBI.C
Author 	 : P. Holy KK Systems Ltd  peter@kksystems.com
Tab setting for editor/printer: 2

MODBUS to Eurotherm Bisync converter for KD485-PROG.

Copyright 1997 KK Systems Ltd

MOD RECORD

1/5/97				Created.
20/5/97		v1	First working & tested & delivered version.
21/5/97		v2	GID & UID transmission fixed. WORKING.
24/5/97		v3	MODBUS Slave ID bit 0 selects direction.
19/7/97		v4	MB regs saved in EEPROM, GID/UID uppercase, MB response
							includes status.
28/7/97		v5	MB ops disabled around EEPROM save.
03/12/97  v105  debug removed
27/02/98  v106  Link formats added
22/12/98  v107  More optimisation, plus abort if another modbus request
                received before responding to previous request.
04/01/99  v108  Hex characters now sent as uppercase.
06/02/99  v109  Removed 50ms delay bug when reading parameters. 

>> The "#define VERSION n" must match the above program version <<

The implementation is synchronous, i.e. an incoming MODBUS message generates
a bisync message to the EI device, obtains the device's response, and returns
a response to the MODBUS master. There is no communication with the EI device
in the absence of MODBUS traffic. Communication with the EI device occurs
only on accesses to the MODBUS parameter register (1-1000).

The MODBUS side is implemented using the KD485 MODBUS Library.
The EI/BISYNC side is implemented with the program below.

See the file EBI.TXT for details of register mapping etc.

When the EI device is on a TWO-wire RS485 port, the RS485 interface must
NOT receive its own transmission. On the KD485-PROG this means the HDX
jumper must be IN. This is because we flush our input queue before we send
the command to the EI device (not after) to avoid any critical timing with
devices which respond very quickly.


CONVERTER DIRECTION

The converter direction is determined from the MODBUS Slave ID byte bit 0.
This byte is configured using KDCFG, as described in the documentation.

bit 0 = 0  ->  port 1 = MODBUS   port 2 = BISYNC
bit 0 = 1  ->  port 1 = BISYNC   port 2 = MODBUS


DEBUG STATEMENTS

If the "#define DEBUG TRUE" is not commented-out, various debug data is
sent to the special KD485-PROG debug port. This a 115200 baud transmit-only
software UART, accessed via the function dfprintf, which is like fprintf
except that the 1st parameter (the destination port, 1 or 2) is missing.
You can also directly call dfwritec which is same as fwritec except again
the destination port is not specified.

As far as is known, the emission of the debug statements does not affect the
operation of this converter in any significant way. They can delay response
by some tens of milliseconds.


*/


#include "modbus.h"
#include "kd485.h"

#define VERSION 109				// Version of this program (integer)

//#define DEBUG TRUE			// Uncomment to include debug statements
#define LINK TRUE
//#define SETADDR TRUE	// Uncomment to override user-settable MB address

#define UCH unsigned char
#define UINT int16

#define STX	0x02
#define ETX	0x03
#define EOT 0x04
#define ENQ 0x05
#define ACK 0x06
#define NAK 0x15

#define ACKRESP		1
#define NAKRESP		2
#define BADRESP		3
#define BADPKT		4
#define EOTRESP		5

#define ICTOUT		8
#define OVTOUT		9
#define NORESP		10

#define BUFLEN	50

// Definitions for location of nonvolatile registers within the EEPROM.

#define EE_BASE				0										// base of EE (relative addr!)
#define EE_TOP 				0x7FFF							// top of usable area
#define EE_SAVE_SIZE 	(2*(4000-1001+1))		// # of *bytes* to save
#define EE_REG_BASE 	EE_TOP-EE_SAVE_SIZE	// EE base for saving the regs


// Global variables

char outbuf[BUFLEN];		// for packets sent to EI device, also used by sysdata
char inbuf[BUFLEN];			// for packets from EI device

const char g_hex_table[] =
              {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

// Storage array for registers 40001-46020
// Note: e.g. MODBUS reg 41001 is stored at array address 0.
#define NUMHOLDINGREGS 6020
int16 HoldingRegs[NUMHOLDINGREGS];

// Storage array for exception registers
uint8 ExceptionRegs[8];

// Address of Slave ID byte in KD485 workspace
uint8	slaveid @ 0xff06;

// Address of MODBUS address byte in KD485 workspace
uint8 mbaddress @ 0xff03;

// Converter ports
uint8 mbport;
uint8 eiport;

// Request flag for EEPROM Save
bool	ee_save_req;

uint16 stp;	// unused


// prototype
void dfwritec( char * buffer, int length );


// Debug function which transmits data to a 115200 baud software UART.
// Like fprintf but you don't specify destination port.
// Maximum # of chars output is 100!

void dprintf(
	char * t,
	...
	)
{
	va_list a;
	int length;
	char buffer[100];
	va_start(a,t);
	length=vsprintf(buffer,t,a);
	va_end(a);
	dfwritec(buffer,length);
}


// Debug function outputs buffer contents, in hex and ascii

void printbuf ( char * buffer, int length )
{
	int j;
	for (j=0;j<length;j++)
		dprintf( "%02x ", (UCH) buffer[j]);
	dprintf( "\r\n");
	for (j=0;j<length;j++)
	{
		if isprint(buffer[j])
			dprintf( " %c ", (UCH) buffer[j]);
		else
			dprintf( "-- ");
	}
}




/*

Function which processes a MODBUS Read command. It generates an EI bisync
Read Command, obtains the EI device's response, and updates MODBUS
registers accordingly

*/

bool RegsRead ( uint16 regstart, uint16 numregs )
{

	uint16 i,j,pktlen,ix,mne,add,dpf,timeout,ictimeout,int1,int2,int3;
	int16 hexvalue,sprm;
	char ch,bcc;
	bool gotch;
	bool badvalue = FALSE;
	bool hexnum = FALSE;
	bool nsign = FALSE;

 	#if DEBUG
	dprintf(  "Register Read : Start reg %d #regs %d\r\n", regstart+1, numregs);
	#endif

	/*
	Check if this is an access from 1-1000 in which case we talk to EI device.
	Accesses to all other registers are handled my the MODBUS library and we
	don't need to do anything to process those.
	Note: 1-1000 corresponds to holdingregs index 0-999.
	Note: the 1-1000 check is done only on the *first* register (of a
	potential multi-register write) because block writes to 1-1000
	are prohibited anyway.
	*/


	if ( regstart < 1000 )
	{

		// From the register set, pick up data needed for the EI packet

		mne = HoldingRegs[regstart+1000];		// mnemonic

		add = HoldingRegs[regstart+2000];		// address

		dpf = HoldingRegs[regstart+3000];		// d.p. / hex format

	#if DEBUG
		dprintf( "Read of a parameter at %d, value = %d\r\n",regstart+1,HoldingRegs[regstart]);
		dprintf( "mnemonic  reg (%4d) is 0x%04x\r\n",regstart+1000+1,mne );
		dprintf( "address   reg (%4d) is 0x%04x\r\n",regstart+2000+1,add );
		dprintf( "dp/format reg (%4d) is 0x%04x\r\n",regstart+3000+1,dpf );
		dprintf( "status    reg (%4d) is 0x%04x\r\n",regstart+5000+1,HoldingRegs[regstart+5000]);
	#endif

	#if DEBUG2
		// Fill outbuf & inbuf with non-digit data
		for (j=0;j<BUFLEN;j++)
		{
			outbuf[j]='?';
			inbuf[j]='?';
		}
    #endif

		// Assemble message to EI device, in outbuf

		outbuf[0]=EOT;

		// Do GID, from high nibble of LSB of address reg
		i = ( add >> 4 ) & 0x000F;
//		sprintf ( &outbuf[1], "%1X", i );
        outbuf[1] = g_hex_table[i];
		outbuf[2] = outbuf[1];

		// Do UID, from low nibble of LSB of address reg
		i = add & 0x000F;
//		sprintf ( &outbuf[3], "%1X", i );
        outbuf[3] = g_hex_table[i];
		outbuf[4] = outbuf[3];

		ix=5;		// use ix from now on as outbuf index, to enable C0 insertion

		// If MSB of address <> 0 then insert that byte as C0

		if ( add >> 8 )
			outbuf[ix++] = add >> 8;

		// Do 1st mnemonic char C1, from MSB of mnemonic reg
		outbuf[ix++] = mne >> 8;
		// Do 2nd mnemonic char C2, from LSB of mnemonic reg
		outbuf[ix++] = mne & 0x00FF;

		// Being an enquiry, there is no data to send, so put in just the ENQ

		outbuf[ix]=ENQ;

		pktlen = ix+1;			// # of bytes in outbuf

	#if DEBUG
		dprintf( "tx buffer=\r\n");
		printbuf( outbuf, pktlen );
	#endif

		// Now output the buffer to the EI device

		ipqclear( eiport );		// flush input queue first

		drvctrl( eiport, 1 );	// enable 485 driver

		fwritec( eiport, outbuf, pktlen );	// output the buffer

		// wait for output queue to empty and then turn off 485 driver

		while ( opqcount (eiport) );
		wait_485(eiport,0);

		/*
		Get the response packet, with a timeout.
		readtimer(0) is the overall timeout, readtimer(1) is the interchar
		timeout.
		*/

		timeout = HoldingRegs[6000];
		ictimeout = HoldingRegs[6001];

		loadtimer(0,timeout);
		gotch=FALSE;
		ch=0;

		/*
		Discard all data until STX or EOT arrives, or until timeout.
		This is always a good idea with 485 comms.
		A STX is the start of a normal packet, while EOT indicates a bad
		mnemonic, and is followed by nothing else.
		*/

		while ( readtimer(0) && !gotch )
			if (ipqcount( eiport ))
			{
				ch=fgetc( eiport );
				if ( (ch == STX) || (ch == EOT) )
				{
					gotch=TRUE;
				}
			}

		// If nothing valid recd then load that status & exit - no more to do

		if ( !gotch )
		{
			HoldingRegs[regstart+5000] = NORESP;
		#if DEBUG
			dprintf( "\r\nNORESP\r\n");
		#endif
			return FALSE;
		}

		// If EOT then load that status and exit - no more to do

		if ( gotch && (ch == EOT) )
		{
			HoldingRegs[regstart+5000] = EOTRESP;
		#if DEBUG
			dprintf( "\r\nEOT recd\r\n");
		#endif
			return FALSE;
		}

		// Get rest of packet, until ETX, or either timeout.

		loadtimer( 1, ictimeout );
		inbuf[0]=STX;
		ix=1;
		ch=0;

		do
		{
			if (ipqcount( eiport ))
			{
				ch=fgetc( eiport );
				inbuf[ix++]=ch;
				if (ix>=(BUFLEN-1))	// just in case we get STX followed by
					ix--;							// a huge amount of garbage
				loadtimer( 1, ictimeout );
			}
		}
		while ( (ch!=ETX) && readtimer(0) && readtimer(1) );

		// Get one more char, should be BCC. If we have had timeout above,
		// this gets skipped.

		while ( readtimer(0) && readtimer(1) )
		{
			if (ipqcount( eiport ))
            {
				inbuf[ix++]=fgetc( eiport );
                break;  // exit loop
            }
		}

		// At this point, we either have recd ETX, or had a timeout

		if ( ch != ETX )
		{

			// Must have had a timeout, one or the other

			if ( readtimer(1)==0 )
			{
				HoldingRegs[regstart+5000] = ICTOUT;
			#if DEBUG
				dprintf( "\r\nICTOUT\r\n");
			#endif
				return FALSE;
			}

			if ( readtimer(0)==0 )
			{
				HoldingRegs[regstart+5000] = OVTOUT;
			#if DEBUG
				dprintf( "\r\nOVTOUT\r\n");
			#endif
				return FALSE;
			}

		}

	#if DEBUG
		dprintf( "\r\nrx buffer=\r\n");
		printbuf( inbuf, ix );
	#endif

		// Check bcc, from 1st char after STX, up to & incl ETX

		bcc=inbuf[1];		// inbuf[1] is 1st char after STX
		ix=2;
		do
		{
			bcc = bcc ^ inbuf[ix++];
		}
		while ( inbuf[ix-1] != ETX );

		if ( inbuf[ix] != bcc )
		{
			HoldingRegs[regstart+5000] = BADRESP;
		#if DEBUG
			dprintf( "\r\nbad bcc, recd=%02x, computed=%02x\r\n",inbuf[ix],bcc);
		#endif
			return FALSE;
		}

		// wipe checksum, else it messes up trailing zero insertion code below
		// (if it happens to be a decimal digit!)

		inbuf[ix]=0;

		/*

		Extract the number from the packet. The packet in inbuf is:
		STX [C0] C1 C2 D1 ... DN ETX BCC
		The numeric field d1..dn is interpreted according to the current
		d.p. specification register, and could turn out to be too big for it.
		The field d1..dn could look like any of the following examples, all of
		which need to be null-terminated for atof() to work, and others need
		additional fixes:

		32767			OK
		-32767		OK
		32767.		needs trailing d.p. removed
		32.767		OK
		0.32767		OK
		.32767		needs a "0" prefix
		-.32767		needs a "-0" prefix in place of the "-"

		The above prefixes are done simply by overwriting the C2 digit. There
		are a few chars before & after the number and these are not useful.

		Actually, we cannot use atof() because it cannot convert *precisely*
		from e.g. "0.00123" to 123 (after multiplication by 100000) but there
		is no harm in leaving these corrections in the code; they make things
		easier.

		Also we have the hex formats, all of which end with the ETX char:

		>A
		>A3
		>A3B
		>A3B8

		*/

		// Make ix -> 1st char of #

		ix=3;							// c1 c2
		if ( add >> 8 )		// c0 c1 c2
			ix++;

		// In case of .xxxxx, we insert 0 before the leading d.p.

		if (inbuf[ix]=='.')
			inbuf[--ix]='0';				// .xxxxx becomes 0.xxxxx
		else
			if ( (inbuf[ix]=='-') && (inbuf[ix+1]=='.') )
			{
				inbuf[--ix]='-';			// -.xxxxx becomes -0.xxxxx
				inbuf[ix+1]='0';
			}

		// Null-terminate the string (required for atof/xtoi) etc

		i=ix;
		while (inbuf[i++]!=ETX) ;		// scan forward to get i -> ETX
		i--;

		if (inbuf[i-1]=='.')	// if last char before ETX is a '.' (xxxxx.)
			inbuf[i-1]=0;				// then put a null over the d.p.
		else
			inbuf[i]=0;					// else put a null over the ETX

		// Extract the number

		if ( inbuf[ix] == '>' )
		{

			// HEX number

			hexnum= TRUE;
			hexvalue = xtoi( (char*) &inbuf[ix+1] );	// +1 to skip ">"
		#if DEBUG
			dprintf( "\r\nhvalue at index=%d is %04x\r\n",ix+1,hexvalue);
		#endif

		}
		else
		{

			/*

			DECIMAL number.

			We cannot use atof() for reasons above, so we use integers. We
			split the number into 3 integer parts, e.g.

			"123.45" splits into

			int1	123
			int2	45

			Next, we generate a "value for MODBUS" (which is always an integer)
			which is int3 and is done as follows:

			int3 = int2 + ( int1 * 10^dpf )

			where dpf is the decimal point spec 0-5, i.e. the # of digits after
			the d.p., e.g. for 123.45 and a d.p. spec of 2 ("xxx.xx"):

			int3 = 45 + 123 * 10^2
			int3 = 45 + 12300 = 12345

			We can detect overflow by checking int1 only; with a d.p. spec of
			2 ("xxx.xx") it must be less than 1000.

			If there was a - sign, we negate int3 at the end.

			The number is now at inbuf[ix] and is null-terminated at the end.

			*/

			int1=0;
			int2=0;
			int3=0;

#if LINK
            // deal with sign
            if((dpf >= 10) && (dpf <= 14))
            {
               // LINK / TCS fixed format, so '-' has replace '.'
			   for ( i=ix; inbuf[i] != '\0'; i++ )
               {
			      if( inbuf[i] == '-' )
			      {
				    inbuf[i]='.';
			        nsign=TRUE;
                    break;
 			      }
               }

               /* re-map to standard type */
               dpf -= 10;
            }
#endif /* LINK */
			// ignore a - sign for conversion, but make a note of it
			if (inbuf[ix]=='-')
			{
			  nsign=TRUE;
			 ix++;
			}

			// find d.p. and put a null over it (for atoi)
			// - only search up to a null, in case there is no d.p., e.g. "12345"

			i=ix;
			while ( ( inbuf[i] ) && ( inbuf[i] != '.' ) )
				i++;

			// Now i -> d.p., or the existing null

			// If d.p. found, load the numbers before/after it into int1/int2

			if ( inbuf[i] == '.' )
			{

				// the part before the d.p. - easy

				inbuf[i++]=0;
				int1 = atoi( &inbuf[ix] );

				// The part after the d.p. - not so easy because one can have e.g.
				// 123.4 or 123.45, both valid with a dpf of 2 ("xxx.xx") !!
				// The part after the d.p. starts at inbuf[i].

				// We add trailing zeroes to equal the dpf, e.g. if dpf=2 then 123.4
				// becomes 123.40. Then we null-terminate the number, for atoi.

				if (dpf)		// if dpf=0 we ignore anything after the d.p.
				{

					j = dpf;
					if ( j > 5 )	// Prevent huge dpf values crashing everything
						j=5;

					while ( j-- )
						if ( !isdigit(inbuf[i++]) )
							inbuf[i-1]='0';

					inbuf[i]=0;
					int2 = atoi( &inbuf[i-dpf] );

				}

			}
			else
			{
				// no d.p. found, simple
				int1 = atoi( &inbuf[ix] );
				int2=0;
			}

			// All the above stuff is just "int3 = atof( inbuf[ix] )" !!

		}

		/*

		Now we have int1,int2 (if decimal#) or hexnum (if hex#).

		Convert the received number into an integer which can be returned to
		the MODBUS parameter register (reg 1-1000).
		We use the D.P. format value to determine what to multiply the number
		by. If the user has configured the correct d.p. format for the range
		of the parameter in question, it is a simple case of e.g.
		123.45 (returned by the EI device)
		with d.p. spec of 2 (xxx.xx)
		being returned as
		12345 (i.e. multiplied by 100)
		in the MODBUS register.
		However, if the user has set-up the wrong d.p. format, e.g. xxxx.x
		then 1234 will be returned, which is wrong. This is not detected.
		If the user has set-up a d.p. format of xx.xxx then there will be an
		overflow since 123.45*1000=123450 will not fit into the 16-bit integer.

		Even with a d.p. spec of 0 (xxxxx) overflow is possible if the number
		returned is > 32767. This should never happen with an EI device.

		Hex numbers are not modified, but overflow is still possible, e.g. the
		number >3AB cannot be returned via a d.p. spec of 7 (>XX).
		Hex numbers are unsigned so can be up to 0xFFFF.

		In summary, the action below is

		0 XXXXX.	leave number as is
		1 XXXX.X	*10
		2 XXX.XX	*100
		3 XX.XXX	*1000
		4 X.XXXX	*10000
		5 .XXXXX	*100000
		6 >X			leave number as is
		7 >XX			leave number as is
		8 >XXX		leave number as is
		9 >XXXX		leave number as is

		*/

		switch (dpf)
		{

		 	case 0 :	if ( ( int1 < 32767 ) && !hexnum )
									int3 = int1;		// int2 should be 0 if no d.p.
								else
									badvalue = TRUE;
							 	break;

		 	case 1 : 	if ( ( int1 < 3277 ) && !hexnum )
									int3 = int2 + int1*10;
								else
									badvalue = TRUE;
							 	break;

		 	case 2 : 	if ( ( int1 < 328 ) && !hexnum )
									int3 = int2 + int1*100;
								else
									badvalue = TRUE;
							 	break;

		 	case 3 : 	if ( ( int1 < 33 ) && !hexnum )
									int3 = int2 + int1*1000;
								else
									badvalue = TRUE;
							 	break;

		 	case 4 : 	if ( ( int1 < 4 ) && !hexnum )
									int3 = int2 + int1*10000;
								else
									badvalue = TRUE;
							 	break;

		 	case 5 :	if ( ( int1 < 1 ) &&
									 ( int2 < 32768 ) &&	// catch e.g. ".40000"
									   !hexnum )
									int3 = int2;
								else
									badvalue = TRUE;
							 	break;

			case 6 : 	if ( ( hexvalue < 16 ) && hexnum )
									sprm = hexvalue;
								else
									badvalue = TRUE;
								break;

			case 7 : 	if ( ( hexvalue < 256 ) && hexnum )
									sprm = hexvalue;
								else
									badvalue = TRUE;
								break;

			case 8 : 	if ( ( hexvalue < 4096 ) && hexnum )
									sprm = hexvalue;
								else
									badvalue = TRUE;
								break;

			case 9 : 	if ( hexnum )
									sprm = hexvalue;
								break;

			default : badvalue = TRUE;
								break;

		}

		if ( !hexnum )
		{
			if ( nsign )
				sprm = - (int16) int3;
			else
				sprm = (int16) int3;
		}

		// Load the MODBUS parameter & status registers

		if ( !badvalue )
		{
		#if DEBUG
			dprintf( "\r\nint1,int2,int3 is %d,%d,%d\r\n",int1,int2,int3);
			dprintf( "ACKRESP, sprm=%d, %04xh, hexnum=%d\r\n",sprm,(UINT)sprm,hexnum);
		#endif
			HoldingRegs[regstart] = sprm;
			HoldingRegs[regstart+5000] = ACKRESP;
		}
		else
		{
		#if DEBUG
			dprintf( "BADPKT\r\n");
		#endif
			HoldingRegs[regstart+5000] = BADPKT;
		}

	}

    // check if another request received
    if (ipqcount( mbport ))
      // abort this response
      return TRUE;


	// If any errors, return 'nack' else return as normal

	if ( badvalue )
	{
		ExceptResponse(NEGATIVE_ACKNOWLEDGE);
		return TRUE;
	}
	else
		return FALSE;

}


/*

Function which processes a MODBUS Write command. It generates an EI bisync
Write Command, obtains the EI device's response, and updates MODBUS
registers accordingly

*/

bool RegsWrite ( uint16 regstart, uint16 numregs )
{

	uint16 i,j,pktlen,ix,mne,add,dpf,timeout;
	float fprm;
	bool badvalue = FALSE;
	bool gotch;
	int16 prmlen=0;
	int16 prm;
	char ch,bcc;

 #if DEBUG
	dprintf(  "Register Write : Start reg %d #regs %d\r\n", regstart+1, numregs);
	dprintf(  "Values written are (hex):\r\n" );
	for ( i=0; i<numregs; i++ )
		dprintf("%02x ",HoldingRegs[i+regstart]);
	dprintf("\r\n");
 #endif

	/*
	Check if this is a write to register 6020. If so, we request an EEPROM save
	and exit right here, returning 'success' to MB master.
	*/

	if ( ( regstart == 6019 ) && ( HoldingRegs[regstart] == 0x55a5 ) )
	{

	 #if DEBUG
		dprintf( "EE Save request\r\n" );
	 #endif

		ee_save_req = TRUE;

		return FALSE;

	}

	/*
	Check if this is an access from 1-1000 in which case we talk to EI device.
	Accesses to all other registers are handled my the MODBUS library and we
	don't need to do anything to process those.
	Note: 1-1000 corresponds to holdingregs index 0-999.
	Note: the 1-1000 check is done only on the *first* register (of a
	potential multi-register write) because block writes to 1-1000
	are prohibited anyway.
	*/

	if ( regstart < 1000 )
	{

		// From the register set, pick up data needed for the EI packet

		prm = HoldingRegs[regstart];				// parameter value (signed!)
		fprm = prm;													// and its floating copy

		mne = HoldingRegs[regstart+1000];		// mnemonic

		add = HoldingRegs[regstart+2000];		// address

		dpf = HoldingRegs[regstart+3000];		// d.p. / hex format

	#if DEBUG
		dprintf( "Write to a parameter at %d, value = %d\r\n",regstart+1,HoldingRegs[regstart]);
		dprintf( "parameter reg (%4d) is 0x%04x\r\n",regstart     +1,prm );
		dprintf( "mnemonic  reg (%4d) is 0x%04x\r\n",regstart+1000+1,mne );
		dprintf( "address   reg (%4d) is 0x%04x\r\n",regstart+2000+1,add );
		dprintf( "dp/format reg (%4d) is 0x%04x\r\n",regstart+3000+1,dpf );
		dprintf( "status    reg (%4d) is 0x%04x\r\n",regstart+5000+1,HoldingRegs[regstart+5000]);
	#endif

		// Fill outbuf with dummy data - makes debugging easier
 	#if DEBUG
		for (j=0;j<30;j++)
			outbuf[j]='?';
    #endif

		// Assemble message to EI device, in outbuf

		outbuf[0]=EOT;

		// Do GID, from high nibble of LSB of address reg
		i = ( add >> 4 ) & 0x000F;
//		sprintf ( &outbuf[1], "%1X", i );
        outbuf[1] = g_hex_table[i];
		outbuf[2] = outbuf[1];

		// Do UID, from low nibble of LSB of address reg
		i = add & 0x000F;
//		sprintf ( &outbuf[3], "%1X", i );
        outbuf[3] = g_hex_table[i];
		outbuf[4] = outbuf[3];

		outbuf[5]=STX;

		ix=6;		// use ix from now on as outbuf index, to enable C0 insertion

		// If MSB of address <> 0 then insert that byte as C0

		if ( add >> 8 )
			outbuf[ix++] = add >> 8;

		// Do 1st mnemonic char C1, from MSB of mnemonic reg
		outbuf[ix++] = mne >> 8;
		// Do 2nd mnemonic char C2, from LSB of mnemonic reg
		outbuf[ix++] = mne & 0x00FF;

		/*
		Do the data field, from the parameter register.
		We use the D.P. format value to format the parameter into a decimal
		string. If we find an invalid d.p. format value we do nothing.
		Now ix=8|9 if C0 not present | present.
		*/

		switch (dpf)
		{

		 	case 0 : 	prmlen=sprintf( &outbuf[ix], "%d.", prm );		// [-]xxxxx.
							 	break;

		 	case 1 : 	prmlen=sprintf( &outbuf[ix], "%1.1f", fprm/10 );	// [-]xxxx.x
							 	break;

		 	case 2 : 	prmlen=sprintf( &outbuf[ix], "%1.2f", fprm/100 );	// [-]xxx.xx
							 	break;

		 	case 3 : 	prmlen=sprintf( &outbuf[ix], "%1.3f", fprm/1000 );	// [-]xx.xxx
							 	break;

		 	case 4 : 	prmlen=sprintf( &outbuf[ix], "%1.4f", fprm/10000 );	// [-]x.xxxx
							 	break;

		 	case 5 : 	if (prm>=0)
									prmlen=sprintf( &outbuf[ix], ".%d", prm );	// .xxxxx
								else
									prmlen=sprintf( &outbuf[ix], "-.%d", abs(prm) );	// -.xxxxx
							 	break;

			case 6 : 	if ( prm < 0x10 )
		 			 				prmlen=sprintf( &outbuf[ix], ">%1X", (UINT)prm );	// >x
								else
									badvalue=TRUE;
								break;

			case 7 : 	if ( prm < 0x100 )
		 							prmlen=sprintf( &outbuf[ix], ">%02X", (UINT)prm );	// >xx
								else
									badvalue=TRUE;
								break;

			case 8 : 	if ( prm < 0x1000 )
		 							prmlen=sprintf( &outbuf[ix], ">%03X", (UINT)prm );	// >xxx
								else
									badvalue=TRUE;
								break;

			case 9 : 	prmlen=sprintf( &outbuf[ix], ">%04X", (UINT)prm );	// >xxxx
								break;

#if LINK
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
                       sprintf( &outbuf[ix], "%04d", abs(prm));
                       i = ix + 4;
                       j = i - (dpf - 10);
                       
                       while( i > j)
                       {
                         outbuf[i] = outbuf[i - 1];
                         i--;
                       }

                       if( prm >= 0 )
                         outbuf[i] = '.';
                       else 
                         outbuf[i] = '-';
                          
                       prmlen = 5;
                       break;
#endif /* LINK */
                       
                                          
			default : badvalue = TRUE;
								break;

		}

		ix=ix+prmlen;			// inc ptr->buffer by # of chars from sprintf above

		outbuf[ix]=ETX;			// put in ETX

		// calculate checksum, from char after STX, to ETX, inclusive

		bcc=0;
		ix=6;		// outbuf[5] is STX
		do
		{
			bcc = bcc ^ outbuf[ix++];
		}
		while ( outbuf[ix-1] != ETX );

		outbuf[ix]=bcc;			// put in checksum, ix now -> BCC byte

		pktlen = ix+1;			// # of bytes in outbuf

	#if DEBUG
		printbuf( outbuf, pktlen );
	#endif

		// Now output the buffer (if valid) to the EI device

		if (badvalue == FALSE)
		{

			ipqclear( eiport );		// flush input queue first

			drvctrl( eiport, 1 );	// enable 485 driver

			fwritec( eiport, outbuf, pktlen );	// output the buffer

			// wait for output queue to empty and then turn off 485 driver

			while ( opqcount (eiport) );
			wait_485(eiport,0);

			// Get the response, ACK or NAK or nothing, with a timeout

			timeout = HoldingRegs[6000];
			loadtimer(0,timeout);
			ch='x';
			gotch=FALSE;

			while ( readtimer(0) && !gotch )
				if (ipqcount( eiport ))
				{
					ch=fgetc( eiport );
					gotch=TRUE;
				}

			if ( gotch )
			{
				switch (ch)
				{
					case ACK :	HoldingRegs[regstart+5000] = ACKRESP;
											break;
					case NAK :	HoldingRegs[regstart+5000] = NAKRESP;
											break;
					default  :	HoldingRegs[regstart+5000] = BADRESP;	// garbage resp
											break;
				}
			}
			else
			{
				HoldingRegs[regstart+5000] = NORESP;
			}

		}
		else
		{
			HoldingRegs[regstart+5000] = BADPKT;
		}

	#if DEBUG
		dprintf( "\r\nch=%02x, status code=%d\r\n",(UCH)ch,HoldingRegs[regstart+5000]);
	#endif

	}

    // check if another request received
    if (ipqcount( mbport ))
      // abort this response
      return TRUE;

	// If any errors, return 'nack' else return as normal
	if ( badvalue )
	{
		ExceptResponse(NEGATIVE_ACKNOWLEDGE);
		return TRUE;
	}
	else
		return FALSE;

}


// These are the routines called by the MODBUS library when a MODBUS message
// is received.

// **** READ ****

bool HoldRegsRead(uint16 startreg, uint16 nregs)
{
	return RegsRead( startreg, nregs );
}

// **** WRITE ****

bool HoldRegsWrite(uint16 startreg, uint16 nregs)
{
	return RegsWrite( startreg, nregs );
}


bool ExRead(uint16 startreg, uint16 nregs)
{
	dprintf(  "Exception Register Read : Start reg %d #regs %d\r\n", startreg+1, nregs);
	return FALSE;
}

void Unknown(uint16 funcnum, uint8 *mess, uint16 messlen)
{
	dprintf(  "Unknown : Function %d Message Length %d\r\n", funcnum, messlen);
}


int main()
{

	uint16 u;

	ee_save_req = FALSE;

// Load nonvolatile MODBUS registers (1001-4000) from EEPROM

	#if DEBUG
	dprintf("\r\n\r\nLoading registers from EEPROM image\r\n");
	dprintf(  "From %04x to %04x size %04x\r\n", (unsigned int)EE_REG_BASE+0x4000, &HoldingRegs[1000], EE_SAVE_SIZE );
	#endif

	// Note that we have to specify the *absolute* address for the EEPROM here
	// hence the +0x4000 which is the start of the EEPROM.

	memcpy( &HoldingRegs[1000], (char*) EE_REG_BASE+0x4000, EE_SAVE_SIZE );

// Initialise any MODBUS registers which are user-optional

	// Timeouts

	HoldingRegs[6000]=100;		// 6001: Response timeout
	HoldingRegs[6001]=50;			// 6002: Interchar response timeout

	// KD485 serial number

	sysdata( outbuf );
	u = (UCH)outbuf[15] + ( 256 * (UCH)outbuf[14] );
  HoldingRegs[6002]=u;

	#if DEBUG
	dprintf("KD485 version %d.%d%c\r\n",(UCH)outbuf[0],(UCH)outbuf[1],outbuf[2]);
	dprintf("firmware date %s\r\n",&outbuf[3]);
	dprintf("%d ports\r\n",(UCH)outbuf[13]);
	dprintf("Serial number %u\r\n",u);
	#endif

	// Converter program version

  HoldingRegs[6003]=VERSION;
	#if DEBUG
	dprintf("Version= %u\r\n",VERSION);
	#endif


// Get converter direction
	mbport = 1 + (slaveid & 0x01);
	if (mbport==1)
		eiport=2;
	else
		eiport=1;

	#if DEBUG
	dprintf("mbport=%d, eiport=%d\r\n", mbport, eiport );
	#endif


// Configure Modbus, if the default port settings are to be changed
// it must be done before this line

	#if USESLAVEID
	if ( SetPort(cvdir+1) < 0 )
	#else
	if ( SetPort(mbport) < 0 )
	#endif
	{
		#if DEBUG
		dprintf(  "Failed to set Port %d for MODBUS\r\n", mbport );
		#endif
   	exit();
	}

// Set the slave address to 1 - this overrides any KDCFG config for this
	#if SETADDR
		#if DEBUG
		dprintf(  "Port Set %d\r\n", mbport);
		dprintf(  "SetSlaveAddress(1)\r\n");
		#endif
		SetSlaveAddress(1);
	#else
		#if DEBUG
		dprintf( "MODBUS address taken from Mode 2 Address: %d\r\n", mbaddress );
		#endif
	#endif


// Set up registers and callbacks for all register types

	// Tell the MODBUS library how many holding regs we have & where
	#if DEBUG
	dprintf(  "SetHoldingRegs\r\n");
	#endif
	SetHoldingRegs(HoldingRegs, NUMHOLDINGREGS);

	// This supports MODBUS functions 3,23
	#if DEBUG
	dprintf(  "SetHoldRegsReadCallback\r\n");
	#endif
	SetHoldRegsReadCallback(HoldRegsRead);

	// This supports MODBUS functions 6,16,22,23
	#if DEBUG
	dprintf(  "SetHoldRegsWriteCallback\r\n");
	#endif
	SetHoldRegsWriteCallback(HoldRegsWrite);

	// Not used
	#if DEBUG
	dprintf(  "SetExceptionRegisters\r\n");
	#endif
	SetExceptionRegisters(ExceptionRegs);

	// Not used
	#if DEBUG
	dprintf(  "SetExReadCallback\r\n");
	#endif
	SetExReadCallback(ExRead);

	// Should never get called
	#if DEBUG
	dprintf(  "SetUnknownCallback\r\n");
	#endif
	SetUnknownCallback(Unknown);

	// Start the modbus library
	#if DEBUG
	dprintf(  "Starting MODBUS\r\n");
	#endif
	Start();


// Loop forever while the modbus library interacts with the master

	/*
	It is here we would put in any async code, e.g. for caching some
	parameters, but we have to be careful with non-re-entrant functions
	like wait_485() etc.
	Also, here we put any code which cannot run inside the callbacks for
	whatever reason.
	*/


	for(;;)
	{

		/*
	
		Function which copies registers 1001-4000 from the reg array to EEPROM.
		The KD485 LED is switched ON during the programming.
		Can take a few seconds.
		This is done here, rather than inside the write callback routine,
		because we want to return a response to the MODBUS
		master quickly, and then allow it to examine the programming status
		register to see how it is doing.

		*/

		if ( ee_save_req )
		{

			ee_save_req = FALSE;		// clear the request flag

			HoldingRegs[6019] = 0;	// indicate 'programming in progress'

		 	#if DEBUG
			dprintf(  "EE save started from %04x to %04x size %04x\r\n", &HoldingRegs[1000], (unsigned int)EE_REG_BASE+0x4000, EE_SAVE_SIZE );
			loadtimer(0,10000);
			#endif

			uled(TRUE);		// LED ON

			Stop();	// One cannot write to the EEPROM while the MB library is active

			u = nvwrblku ( (char*) &HoldingRegs[1000], EE_REG_BASE, EE_SAVE_SIZE );

			Start();

			uled(FALSE);	// LED OFF

		 	#if DEBUG
			dprintf(  "EE save finished, ret=%d, duration=%dms\r\n", u, 10000-readtimer(0) );
			#endif

			if (u==0)			// if nvwrblku returned 'success' (0)
				u=10;				// then we return 10 to indicate this

			HoldingRegs[6019] = u;	// load result of programming

		}

	}	

}


